@@ -1,4 +1,6 @@
 'use strict'
+/* eslint-disable camelcase */
+/* eslint-disable standard/no-callback-literal */
 // npm install <pkg> <pkg> <pkg>
 //
 // See doc/cli/npm-install.md for more description
@@ -29,7 +31,7 @@ install.usage = usage(
   '\nnpm install <tarball url>' +
   '\nnpm install <git:// url>' +
   '\nnpm install <github username>/<github project>',
-  '[--save|--save-dev|--save-optional] [--save-exact]'
+  '[--save-prod|--save-dev|--save-optional] [--save-exact] [--no-save]'
 )
 
 install.completion = function (opts, cb) {
@@ -98,6 +100,7 @@ var path = require('path')
 // dependencies
 var log = require('npmlog')
 var readPackageTree = require('read-package-tree')
+var readPackageJson = require('read-package-json')
 var chain = require('slide').chain
 var asyncMap = require('slide').asyncMap
 var archy = require('archy')
@@ -105,20 +108,23 @@ var mkdirp = require('mkdirp')
 var rimraf = require('rimraf')
 var iferr = require('iferr')
 var validate = require('aproba')
+var uniq = require('lodash.uniq')
+var Bluebird = require('bluebird')
 
 // npm internal utils
 var npm = require('./npm.js')
 var locker = require('./utils/locker.js')
 var lock = locker.lock
 var unlock = locker.unlock
-var ls = require('./ls.js')
 var parseJSON = require('./utils/parse-json.js')
 var output = require('./utils/output.js')
+var saveMetrics = require('./utils/metrics.js').save
 
 // install specific libraries
 var copyTree = require('./install/copy-tree.js')
 var readShrinkwrap = require('./install/read-shrinkwrap.js')
-var recalculateMetadata = require('./install/deps.js').recalculateMetadata
+var computeMetadata = require('./install/deps.js').computeMetadata
+var prefetchDeps = require('./install/deps.js').prefetchDeps
 var loadDeps = require('./install/deps.js').loadDeps
 var loadDevDeps = require('./install/deps.js').loadDevDeps
 var getAllMetadata = require('./install/deps.js').getAllMetadata
@@ -127,20 +133,23 @@ var loadExtraneous = require('./install/deps.js').loadExtraneous
 var diffTrees = require('./install/diff-trees.js')
 var checkPermissions = require('./install/check-permissions.js')
 var decomposeActions = require('./install/decompose-actions.js')
-var filterInvalidActions = require('./install/filter-invalid-actions.js')
 var validateTree = require('./install/validate-tree.js')
 var validateArgs = require('./install/validate-args.js')
 var saveRequested = require('./install/save.js').saveRequested
+var saveShrinkwrap = require('./install/save.js').saveShrinkwrap
+var audit = require('./install/audit.js')
 var getSaveType = require('./install/save.js').getSaveType
 var doSerialActions = require('./install/actions.js').doSerial
 var doReverseSerialActions = require('./install/actions.js').doReverseSerial
 var doParallelActions = require('./install/actions.js').doParallel
 var doOneAction = require('./install/actions.js').doOne
 var removeObsoleteDep = require('./install/deps.js').removeObsoleteDep
+var removeExtraneous = require('./install/deps.js').removeExtraneous
+var computeVersionSpec = require('./install/deps.js').computeVersionSpec
 var packageId = require('./utils/package-id.js')
 var moduleName = require('./utils/module-name.js')
 var errorMessage = require('./utils/error-message.js')
-var andIgnoreErrors = require('./install/and-ignore-errors.js')
+var isExtraneous = require('./install/is-extraneous.js')
 
 function unlockCB (lockPath, name, cb) {
   validate('SSF', arguments)
@@ -175,8 +184,8 @@ function install (where, args, cb) {
   var globalTop = path.resolve(npm.globalDir, '..')
   if (!where) {
     where = npm.config.get('global')
-          ? globalTop
-          : npm.prefix
+      ? globalTop
+      : npm.prefix
   }
   validate('SAF', [where, args, cb])
   // the /path/to/node_modules/..
@@ -196,11 +205,16 @@ function install (where, args, cb) {
   new Installer(where, dryrun, args).run(cb)
 }
 
-function Installer (where, dryrun, args) {
-  validate('SBA', arguments)
+function Installer (where, dryrun, args, opts) {
+  validate('SBA|SBAO', arguments)
+  if (!opts) opts = {}
   this.where = where
   this.dryrun = dryrun
   this.args = args
+  // fakechildren are children created from the lockfile and lack relationship data
+  // the only exist when the tree does not match the lockfile
+  // this is fine when doing full tree installs/updates but not ok when modifying only
+  // a few deps via `npm install` or `npm uninstall`.
   this.currentTree = null
   this.idealTree = null
   this.differences = []
@@ -208,18 +222,43 @@ function Installer (where, dryrun, args) {
   this.progress = {}
   this.noPackageJsonOk = !!args.length
   this.topLevelLifecycles = !args.length
-  this.npat = npm.config.get('npat')
-  this.dev = npm.config.get('dev') || (!/^prod(uction)?$/.test(npm.config.get('only')) && !npm.config.get('production')) || /^dev(elopment)?$/.test(npm.config.get('only'))
-  this.prod = !/^dev(elopment)?$/.test(npm.config.get('only'))
-  this.rollback = npm.config.get('rollback')
-  this.link = npm.config.get('link')
-  this.global = this.where === path.resolve(npm.globalDir, '..')
+
+  this.autoPrune = npm.config.get('package-lock')
+
+  const dev = npm.config.get('dev')
+  const only = npm.config.get('only')
+  const onlyProd = /^prod(uction)?$/.test(only)
+  const onlyDev = /^dev(elopment)?$/.test(only)
+  const prod = npm.config.get('production')
+  this.dev = opts.dev != null ? opts.dev : dev || (!onlyProd && !prod) || onlyDev
+  this.prod = opts.prod != null ? opts.prod : !onlyDev
+
+  this.packageLockOnly = opts.packageLockOnly != null
+    ? opts.packageLockOnly : npm.config.get('package-lock-only')
+  this.rollback = opts.rollback != null ? opts.rollback : npm.config.get('rollback')
+  this.link = opts.link != null ? opts.link : npm.config.get('link')
+  this.saveOnlyLock = opts.saveOnlyLock
+  this.global = opts.global != null ? opts.global : this.where === path.resolve(npm.globalDir, '..')
+  this.audit = npm.config.get('audit') && !this.global
+  this.started = Date.now()
 }
 Installer.prototype = {}
 
-Installer.prototype.run = function (cb) {
-  validate('F', arguments)
+Installer.prototype.run = function (_cb) {
+  validate('F|', arguments)
 
+  var result
+  var cb
+  if (_cb) {
+    cb = function (err) {
+      saveMetrics(!err)
+      return _cb.apply(this, arguments)
+    }
+  } else {
+    result = new Promise((resolve, reject) => {
+      cb = (err, value) => err ? reject(err) : resolve(value)
+    })
+  }
   // FIXME: This is bad and I should feel bad.
   // lib/install needs to have some way of sharing _limited_
   // state with the things it calls. Passing the object is too
@@ -239,6 +278,11 @@ Installer.prototype.run = function (cb) {
 
   var installSteps = []
   var postInstallSteps = []
+  if (!this.dryrun) {
+    installSteps.push(
+      [this.newTracker(log, 'runTopLevelLifecycles', 2)],
+      [this, this.runPreinstallTopLevelLifecycles])
+  }
   installSteps.push(
     [this.newTracker(log, 'loadCurrentTree', 4)],
     [this, this.loadCurrentTree],
@@ -256,12 +300,15 @@ Installer.prototype.run = function (cb) {
     [this, this.finishTracker, 'generateActionsToTake'],
 
     [this, this.debugActions, 'diffTrees', 'differences'],
-    [this, this.debugActions, 'decomposeActions', 'todo'])
-  if (!this.dryrun) {
-    installSteps.push(
-      [this.newTracker(log, 'runTopLevelLifecycles', 2)],
-      [this, this.runPreinstallTopLevelLifecycles],
+    [this, this.debugActions, 'decomposeActions', 'todo'],
+    [this, this.startAudit]
+  )
 
+  if (this.packageLockOnly) {
+    postInstallSteps.push(
+      [this, this.saveToDependencies])
+  } else if (!this.dryrun) {
+    installSteps.push(
       [this.newTracker(log, 'executeActions', 8)],
       [this, this.executeActions],
       [this, this.finishTracker, 'executeActions'])
@@ -274,28 +321,26 @@ Installer.prototype.run = function (cb) {
       [this, this.commit, staging, this.todo],
 
       [this, this.runPostinstallTopLevelLifecycles],
-      [this, this.finishTracker, 'runTopLevelLifecycles'])
-    if (getSaveType(this.args)) {
+      [this, this.finishTracker, 'runTopLevelLifecycles']
+    )
+    if (getSaveType()) {
       postInstallSteps.push(
+        // this is necessary as we don't fill in `dependencies` and `devDependencies` in deps loaded from shrinkwrap
+        // until after we extract them
+        [this, (next) => { computeMetadata(this.idealTree); next() }],
+        [this, this.pruneIdealTree],
+        [this, this.debugLogicalTree, 'saveTree', 'idealTree'],
         [this, this.saveToDependencies])
     }
   }
   postInstallSteps.push(
+    [this, this.printWarnings],
     [this, this.printInstalled])
 
   var self = this
   chain(installSteps, function (installEr) {
     if (installEr) self.failing = true
     chain(postInstallSteps, function (postInstallEr) {
-      if (self.idealTree) {
-        self.idealTree.warnings.forEach(function (warning) {
-          if (warning.code === 'EPACKAGEJSON' && self.global) return
-          if (warning.code === 'ENOTDIR') return
-          errorMessage(warning).summary.forEach(function (logline) {
-            log.warn.apply(log, logline)
-          })
-        })
-      }
       if (installEr && postInstallEr) {
         var msg = errorMessage(postInstallEr)
         msg.summary.forEach(function (logline) {
@@ -308,12 +353,12 @@ Installer.prototype.run = function (cb) {
       cb(installEr || postInstallEr, self.getInstalledModules(), self.idealTree)
     })
   })
+  return result
 }
 
 Installer.prototype.loadArgMetadata = function (next) {
-  var self = this
-  getAllMetadata(this.args, this.currentTree, process.cwd(), iferr(next, function (args) {
-    self.args = args
+  getAllMetadata(this.args, this.currentTree, process.cwd(), iferr(next, (args) => {
+    this.args = args
     next()
   }))
 }
@@ -322,17 +367,15 @@ Installer.prototype.newTracker = function (tracker, name, size) {
   validate('OS', [tracker, name])
   if (size) validate('N', [size])
   this.progress[name] = tracker.newGroup(name, size)
-  var self = this
   return function (next) {
-    self.progress[name].silly(name, 'Starting')
+    process.emit('time', 'stage:' + name)
     next()
   }
 }
 
 Installer.prototype.finishTracker = function (name, cb) {
   validate('SF', arguments)
-  this.progress[name].silly(name, 'Finishing')
-  this.progress[name].finish()
+  process.emit('timeEnd', 'stage:' + name)
   cb()
 }
 
@@ -345,39 +388,74 @@ Installer.prototype.loadCurrentTree = function (cb) {
   } else {
     todo.push([this, this.readLocalPackageData])
   }
-  todo.push(
-    [this, this.normalizeTree, log.newGroup('normalizeTree')])
+  todo.push([this, this.normalizeCurrentTree])
   chain(todo, cb)
 }
 
+var createNode = require('./install/node.js').create
+var flatNameFromTree = require('./install/flatten-tree.js').flatNameFromTree
+Installer.prototype.normalizeCurrentTree = function (cb) {
+  this.currentTree.isTop = true
+  normalizeTree(this.currentTree)
+  // If the user didn't have a package.json then fill in deps with what was on disk
+  if (this.currentTree.error) {
+    for (let child of this.currentTree.children) {
+      if (!child.fakeChild && isExtraneous(child)) {
+        this.currentTree.package.dependencies[child.package.name] = computeVersionSpec(this.currentTree, child)
+      }
+    }
+  }
+  computeMetadata(this.currentTree)
+  return cb()
+
+  function normalizeTree (tree, seen) {
+    if (!seen) seen = new Set()
+    if (seen.has(tree)) return
+    seen.add(tree)
+    createNode(tree)
+    tree.location = flatNameFromTree(tree)
+    tree.children.forEach((child) => normalizeTree(child, seen))
+  }
+}
+
 Installer.prototype.loadIdealTree = function (cb) {
   validate('F', arguments)
   log.silly('install', 'loadIdealTree')
 
   chain([
-    [this.newTracker(this.progress.loadIdealTree, 'cloneCurrentTree')],
+    [this.newTracker(this.progress.loadIdealTree, 'loadIdealTree:cloneCurrentTree')],
     [this, this.cloneCurrentTreeToIdealTree],
-    [this, this.finishTracker, 'cloneCurrentTree'],
+    [this, this.finishTracker, 'loadIdealTree:cloneCurrentTree'],
 
-    [this.newTracker(this.progress.loadIdealTree, 'loadShrinkwrap')],
+    [this.newTracker(this.progress.loadIdealTree, 'loadIdealTree:loadShrinkwrap')],
     [this, this.loadShrinkwrap],
-    [this, this.finishTracker, 'loadShrinkwrap'],
+    [this, this.finishTracker, 'loadIdealTree:loadShrinkwrap'],
 
-    [this.newTracker(this.progress.loadIdealTree, 'loadAllDepsIntoIdealTree', 10)],
+    [this.newTracker(this.progress.loadIdealTree, 'loadIdealTree:loadAllDepsIntoIdealTree', 10)],
     [this, this.loadAllDepsIntoIdealTree],
-    [this, this.finishTracker, 'loadAllDepsIntoIdealTree'],
-
-    // TODO: Remove this (should no longer be necessary, instead counter productive)
-    [this, function (next) { recalculateMetadata(this.idealTree, log, next) }]
+    [this, this.finishTracker, 'loadIdealTree:loadAllDepsIntoIdealTree'],
+    [this, function (next) { computeMetadata(this.idealTree); next() }],
+    [this, this.pruneIdealTree]
   ], cb)
 }
 
+Installer.prototype.pruneIdealTree = function (cb) {
+  if (!this.idealTree) return cb()
+  // if our lock file didn't have the requires field and there
+  // are any fake children then forgo pruning until we have more info.
+  if (!this.idealTree.hasRequiresFromLock && this.idealTree.children.some((n) => n.fakeChild)) return cb()
+  const toPrune = this.idealTree.children
+    .filter((child) => isExtraneous(child) && (this.autoPrune || child.removing))
+    .map((n) => ({name: moduleName(n)}))
+  return removeExtraneous(toPrune, this.idealTree, cb)
+}
+
 Installer.prototype.loadAllDepsIntoIdealTree = function (cb) {
   validate('F', arguments)
   log.silly('install', 'loadAllDepsIntoIdealTree')
-  var saveDeps = getSaveType(this.args)
+  var saveDeps = getSaveType()
 
-  var cg = this.progress.loadAllDepsIntoIdealTree
+  var cg = this.progress['loadIdealTree:loadAllDepsIntoIdealTree']
   var installNewModules = !!this.args.length
   var steps = []
 
@@ -385,14 +463,14 @@ Installer.prototype.loadAllDepsIntoIdealTree = function (cb) {
     steps.push([validateArgs, this.idealTree, this.args])
     steps.push([loadRequestedDeps, this.args, this.idealTree, saveDeps, cg.newGroup('loadRequestedDeps')])
   } else {
-    if (this.prod) {
-      steps.push(
-        [loadDeps, this.idealTree, cg.newGroup('loadDeps')])
-    }
-    if (this.dev) {
-      steps.push(
-        [loadDevDeps, this.idealTree, cg.newGroup('loadDevDeps')])
-    }
+    const depsToPreload = Object.assign({},
+      this.idealTree.package.devDependencies,
+      this.idealTree.package.dependencies
+    )
+    steps.push(
+      [prefetchDeps, this.idealTree, depsToPreload, cg.newGroup('prefetchDeps')],
+      [loadDeps, this.idealTree, cg.newGroup('loadDeps')],
+      [loadDevDeps, this.idealTree, cg.newGroup('loadDevDeps')])
   }
   steps.push(
     [loadExtraneous.andResolveDeps, this.idealTree, cg.newGroup('loadExtraneous')])
@@ -407,7 +485,6 @@ Installer.prototype.generateActionsToTake = function (cb) {
     [validateTree, this.idealTree, cg.newGroup('validateTree')],
     [diffTrees, this.currentTree, this.idealTree, this.differences, cg.newGroup('diffTrees')],
     [this, this.computeLinked],
-    [filterInvalidActions, this.where, this.differences],
     [checkPermissions, this.differences],
     [decomposeActions, this.differences, this.todo]
   ], cb)
@@ -467,24 +544,20 @@ Installer.prototype.executeActions = function (cb) {
 
   steps.push(
     [doSerialActions, 'global-install', staging, todo, trackLifecycle.newGroup('global-install')],
-    [doParallelActions, 'fetch', staging, todo, cg.newGroup('fetch', 10)],
     [lock, node_modules, '.staging'],
     [rimraf, staging],
-    [mkdirp, staging],
-    [doParallelActions, 'extract', staging, todo, cg.newGroup('extract', 10)],
-    [doParallelActions, 'preinstall', staging, todo, trackLifecycle.newGroup('preinstall')],
-    [doReverseSerialActions, 'remove', staging, todo, cg.newGroup('remove')],
+    [doParallelActions, 'extract', staging, todo, cg.newGroup('extract', 100)],
+    [doReverseSerialActions, 'unbuild', staging, todo, cg.newGroup('unbuild')],
+    [doSerialActions, 'remove', staging, todo, cg.newGroup('remove')],
     [doSerialActions, 'move', staging, todo, cg.newGroup('move')],
     [doSerialActions, 'finalize', staging, todo, cg.newGroup('finalize')],
+    [doParallelActions, 'refresh-package-json', staging, todo, cg.newGroup('refresh-package-json')],
+    [doParallelActions, 'preinstall', staging, todo, trackLifecycle.newGroup('preinstall')],
     [doSerialActions, 'build', staging, todo, trackLifecycle.newGroup('build')],
     [doSerialActions, 'global-link', staging, todo, trackLifecycle.newGroup('global-link')],
     [doParallelActions, 'update-linked', staging, todo, trackLifecycle.newGroup('update-linked')],
     [doSerialActions, 'install', staging, todo, trackLifecycle.newGroup('install')],
     [doSerialActions, 'postinstall', staging, todo, trackLifecycle.newGroup('postinstall')])
-  if (this.npat) {
-    steps.push(
-      [doParallelActions, 'test', staging, todo, trackLifecycle.newGroup('npat')])
-  }
 
   var self = this
   chain(steps, function (er) {
@@ -498,16 +571,15 @@ Installer.prototype.executeActions = function (cb) {
 
 Installer.prototype.rollbackFailedOptional = function (staging, actionsToRun, cb) {
   if (!this.rollback) return cb()
-  var failed = actionsToRun.map(function (action) {
+  var failed = uniq(actionsToRun.map(function (action) {
     return action[1]
   }).filter(function (pkg) {
     return pkg.failed && pkg.rollback
-  })
-  asyncMap(failed, function (pkg, next) {
-    asyncMap(pkg.rollback, function (rollback, done) {
-      rollback(staging, pkg, done)
-    }, next)
-  }, cb)
+  }))
+  var top = this.currentTree && this.currentTree.path
+  Bluebird.map(failed, (pkg) => {
+    return Bluebird.map(pkg.rollback, (rollback) => rollback(top, staging, pkg))
+  }).asCallback(cb)
 }
 
 Installer.prototype.commit = function (staging, actionsToRun, cb) {
@@ -527,13 +599,16 @@ Installer.prototype.runPreinstallTopLevelLifecycles = function (cb) {
   if (this.failing) return cb()
   if (!this.topLevelLifecycles) return cb()
   log.silly('install', 'runPreinstallTopLevelLifecycles')
-  var steps = []
-  var trackLifecycle = this.progress.runTopLevelLifecycles
 
-  steps.push(
-    [doOneAction, 'preinstall', this.idealTree.path, this.idealTree, trackLifecycle.newGroup('preinstall:.')]
-  )
-  chain(steps, cb)
+  readPackageJson(path.join(this.where, 'package.json'), log, false, (err, data) => {
+    if (err) return cb()
+    this.currentTree = createNode({
+      isTop: true,
+      package: data,
+      path: this.where
+    })
+    doOneAction('preinstall', this.where, this.currentTree, log.newGroup('preinstall:.'), cb)
+  })
 }
 
 Installer.prototype.runPostinstallTopLevelLifecycles = function (cb) {
@@ -548,22 +623,32 @@ Installer.prototype.runPostinstallTopLevelLifecycles = function (cb) {
     [doOneAction, 'build', this.idealTree.path, this.idealTree, trackLifecycle.newGroup('build:.')],
     [doOneAction, 'install', this.idealTree.path, this.idealTree, trackLifecycle.newGroup('install:.')],
     [doOneAction, 'postinstall', this.idealTree.path, this.idealTree, trackLifecycle.newGroup('postinstall:.')])
-  if (this.npat) {
-    steps.push(
-      [doOneAction, 'test', this.idealTree.path, this.idealTree, trackLifecycle.newGroup('npat:.')])
-  }
   if (this.dev) {
     steps.push(
-      [doOneAction, 'prepublish', this.idealTree.path, this.idealTree, trackLifecycle.newGroup('prepublish')])
+      [doOneAction, 'prepare', this.idealTree.path, this.idealTree, trackLifecycle.newGroup('prepare')])
   }
   chain(steps, cb)
 }
 
+Installer.prototype.startAudit = function (cb) {
+  if (!this.audit) return cb()
+  this.auditSubmission = Bluebird.try(() => {
+    return audit.generateFromInstall(this.idealTree, this.differences, this.args, this.remove)
+  }).then((auditData) => {
+    return audit.submitForInstallReport(auditData)
+  }).catch(_ => {})
+  cb()
+}
+
 Installer.prototype.saveToDependencies = function (cb) {
   validate('F', arguments)
   if (this.failing) return cb()
   log.silly('install', 'saveToDependencies')
-  saveRequested(this.args, this.idealTree, cb)
+  if (this.saveOnlyLock) {
+    saveShrinkwrap(this.idealTree, cb)
+  } else {
+    saveRequested(this.idealTree, cb)
+  }
 }
 
 Installer.prototype.readGlobalPackageData = function (cb) {
@@ -603,19 +688,13 @@ Installer.prototype.readLocalPackageData = function (cb) {
         return cb(er)
       }
       if (!currentTree.package) currentTree.package = {}
-      if (currentTree.package._shrinkwrap) {
-        self.loadArgMetadata(cb)
-      } else {
-        fs.readFile(path.join(self.where, 'npm-shrinkwrap.json'), function (er, data) {
-          if (er) return self.loadArgMetadata(cb)
-          try {
-            currentTree.package._shrinkwrap = parseJSON(data)
-          } catch (ex) {
-            return cb(ex)
-          }
-          return self.loadArgMetadata(cb)
-        })
-      }
+      readShrinkwrap(currentTree, function (err) {
+        if (err) {
+          cb(err)
+        } else {
+          self.loadArgMetadata(cb)
+        }
+      })
     }))
   }))
 }
@@ -623,6 +702,7 @@ Installer.prototype.readLocalPackageData = function (cb) {
 Installer.prototype.cloneCurrentTreeToIdealTree = function (cb) {
   validate('F', arguments)
   log.silly('install', 'cloneCurrentTreeToIdealTree')
+
   this.idealTree = copyTree(this.currentTree)
   this.idealTree.warnings = []
   cb()
@@ -631,24 +711,9 @@ Installer.prototype.cloneCurrentTreeToIdealTree = function (cb) {
 Installer.prototype.loadShrinkwrap = function (cb) {
   validate('F', arguments)
   log.silly('install', 'loadShrinkwrap')
-  var installNewModules = !!this.args.length
-  if (installNewModules) {
-    readShrinkwrap(this.idealTree, cb)
-  } else {
-    readShrinkwrap.andInflate(this.idealTree, cb)
-  }
-}
-
-Installer.prototype.normalizeTree = function (log, cb) {
-  validate('OF', arguments)
-  log.silly('install', 'normalizeTree')
-  recalculateMetadata(this.currentTree, log, iferr(cb, function (tree) {
-    tree.children.forEach(function (child) {
-      if (child.requiredBy.length === 0) {
-        child.existing = true
-      }
-    })
-    cb(null, tree)
+  readShrinkwrap.andInflate(this.idealTree, iferr(cb, () => {
+    computeMetadata(this.idealTree)
+    cb()
   }))
 }
 
@@ -662,46 +727,227 @@ Installer.prototype.getInstalledModules = function () {
   })
 }
 
+Installer.prototype.printWarnings = function (cb) {
+  if (!this.idealTree) return cb()
+
+  var self = this
+  var warned = false
+  this.idealTree.warnings.forEach(function (warning) {
+    if (warning.code === 'EPACKAGEJSON' && self.global) return
+    if (warning.code === 'ENOTDIR') return
+    warned = true
+    var msg = errorMessage(warning)
+    msg.summary.forEach(function (logline) {
+      log.warn.apply(log, logline)
+    })
+    msg.detail.forEach(function (logline) {
+      log.verbose.apply(log, logline)
+    })
+  })
+  if (warned && log.levels[npm.config.get('loglevel')] <= log.levels.warn) console.error()
+  cb()
+}
+
 Installer.prototype.printInstalled = function (cb) {
   validate('F', arguments)
+  if (this.failing) return cb()
   log.silly('install', 'printInstalled')
-  var self = this
-  this.differences.forEach(function (action) {
+  const diffs = this.differences
+  if (!this.idealTree.error && this.idealTree.removedChildren) {
+    const deps = this.currentTree.package.dependencies || {}
+    const dev = this.currentTree.package.devDependencies || {}
+    this.idealTree.removedChildren.forEach((r) => {
+      if (diffs.some((d) => d[0] === 'remove' && d[1].path === r.path)) return
+      if (!deps[moduleName(r)] && !dev[moduleName(r)]) return
+      diffs.push(['remove', r])
+    })
+  }
+  return Bluebird.try(() => {
+    return this.auditSubmission && this.auditSubmission.catch(() => null)
+  }).then((auditResult) => {
+    // maybe write audit report w/ hash of pjson & shrinkwrap for later reading by `npm audit`
+    if (npm.config.get('json')) {
+      return this.printInstalledForJSON(diffs, auditResult)
+    } else if (npm.config.get('parseable')) {
+      return this.printInstalledForParseable(diffs, auditResult)
+    } else {
+      return this.printInstalledForHuman(diffs, auditResult)
+    }
+  }).asCallback(cb)
+}
+
+Installer.prototype.printInstalledForHuman = function (diffs, auditResult) {
+  var removed = 0
+  var added = 0
+  var updated = 0
+  var moved = 0
+  // Count the number of contributors to packages added, tracking
+  // contributors we've seen, so we can produce a running unique count.
+  var contributors = new Set()
+  diffs.forEach(function (action) {
     var mutation = action[0]
-    var child = action[1]
-    var name = packageId(child)
-    var where = path.relative(self.where, child.path)
+    var pkg = action[1]
+    if (pkg.failed) return
     if (mutation === 'remove') {
-      output('- ' + name + ' ' + where)
+      ++removed
     } else if (mutation === 'move') {
-      var oldWhere = path.relative(self.where, child.fromPath)
-      output(name + ' ' + oldWhere + ' -> ' + where)
+      ++moved
+    } else if (mutation === 'add') {
+      ++added
+      // Count contributors to added packages. Start by combining `author`
+      // and `contributors` data into a single array of contributor-people
+      // for this package.
+      var people = []
+      var meta = pkg.package
+      if (meta.author) people.push(meta.author)
+      if (meta.contributors && Array.isArray(meta.contributors)) {
+        people = people.concat(meta.contributors)
+      }
+      // Make sure a normalized string for every person behind this
+      // package is in `contributors`.
+      people.forEach(function (person) {
+        // Ignore errors from malformed `author` and `contributors`.
+        try {
+          var normalized = normalizePerson(person)
+        } catch (error) {
+          return
+        }
+        if (!contributors.has(normalized)) contributors.add(normalized)
+      })
+    } else if (mutation === 'update' || mutation === 'update-linked') {
+      ++updated
     }
   })
-  var addedOrMoved = this.differences.filter(function (action) {
+  var report = ''
+  if (this.args.length && (added || updated)) {
+    report += this.args.map((p) => {
+      return `+ ${p.name}@${p.version}`
+    }).join('\n') + '\n'
+  }
+  var actions = []
+  if (added) {
+    var action = 'added ' + packages(added)
+    if (contributors.size) action += from(contributors.size)
+    actions.push(action)
+  }
+  if (removed) actions.push('removed ' + packages(removed))
+  if (updated) actions.push('updated ' + packages(updated))
+  if (moved) actions.push('moved ' + packages(moved))
+  if (actions.length === 0) {
+    report += 'up to date'
+  } else if (actions.length === 1) {
+    report += actions[0]
+  } else {
+    var lastAction = actions.pop()
+    report += actions.join(', ') + ' and ' + lastAction
+  }
+  report += ' in ' + ((Date.now() - this.started) / 1000) + 's'
+
+  output(report)
+  return auditResult && audit.printInstallReport(auditResult)
+
+  function packages (num) {
+    return num + ' package' + (num > 1 ? 's' : '')
+  }
+
+  function from (num) {
+    return ' from ' + num + ' contributor' + (num > 1 ? 's' : '')
+  }
+
+  // Values of `author` and elements of `contributors` in `package.json`
+  // files can be e-mail style strings or Objects with `name`, `email,
+  // and `url` String properties.  Convert Objects to Strings so that
+  // we can efficiently keep a set of contributors we have already seen.
+  function normalizePerson (argument) {
+    if (typeof argument === 'string') return argument
+    var returned = ''
+    if (argument.name) returned += argument.name
+    if (argument.email) returned += ' <' + argument.email + '>'
+    if (argument.url) returned += ' (' + argument.email + ')'
+    return returned
+  }
+}
+
+Installer.prototype.printInstalledForJSON = function (diffs, auditResult) {
+  var result = {
+    added: [],
+    removed: [],
+    updated: [],
+    moved: [],
+    failed: [],
+    warnings: [],
+    audit: auditResult,
+    elapsed: Date.now() - this.started
+  }
+  var self = this
+  this.idealTree.warnings.forEach(function (warning) {
+    if (warning.code === 'EPACKAGEJSON' && self.global) return
+    if (warning.code === 'ENOTDIR') return
+    var output = errorMessage(warning)
+    var message = flattenMessage(output.summary)
+    if (output.detail.length) {
+      message += '\n' + flattenMessage(output.detail)
+    }
+    result.warnings.push(message)
+  })
+  diffs.forEach(function (action) {
     var mutation = action[0]
     var child = action[1]
-    return !child.failed && (mutation === 'add' || mutation === 'update')
-  }).map(function (action) {
+    var record = recordAction(action)
+    if (child.failed) {
+      result.failed.push(record)
+    } else if (mutation === 'add') {
+      result.added.push(record)
+    } else if (mutation === 'update' || mutation === 'update-linked') {
+      result.updated.push(record)
+    } else if (mutation === 'move') {
+      result.moved.push(record)
+    } else if (mutation === 'remove') {
+      result.removed.push(record)
+    }
+  })
+  output(JSON.stringify(result, null, 2))
+
+  function flattenMessage (msg) {
+    return msg.map(function (logline) { return logline.slice(1).join(' ') }).join('\n')
+  }
+
+  function recordAction (action) {
+    var mutation = action[0]
     var child = action[1]
-    return child.path
+    var result = {
+      action: mutation,
+      name: moduleName(child),
+      version: child.package && child.package.version,
+      path: child.path
+    }
+    if (mutation === 'move') {
+      result.previousPath = child.fromPath
+    } else if (mutation === 'update') {
+      result.previousVersion = child.oldPkg.package && child.oldPkg.package.version
+    }
+    return result
+  }
+}
+
+Installer.prototype.printInstalledForParseable = function (diffs) {
+  var self = this
+  diffs.forEach(function (action) {
+    var mutation = action[0]
+    var child = action[1]
+    if (mutation === 'move') {
+      var previousPath = path.relative(self.where, child.fromPath)
+    } else if (mutation === 'update') {
+      var previousVersion = child.oldPkg.package && child.oldPkg.package.version
+    }
+    output(
+      mutation + '\t' +
+      moduleName(child) + '\t' +
+      (child.package ? child.package.version : '') + '\t' +
+      (child.path ? path.relative(self.where, child.path) : '') + '\t' +
+      (previousVersion || '') + '\t' +
+      (previousPath || ''))
   })
-  if (!addedOrMoved.length) return cb()
-  // TODO: remove the recalculateMetadata, should not be needed
-  recalculateMetadata(this.idealTree, log, iferr(cb, function (tree) {
-    // These options control both how installs happen AND how `ls` shows output.
-    // Something like `npm install --production` only installs production deps.
-    // By contrast `npm install --production foo` installs `foo` and the
-    // `production` option is ignored. But when it comes time for `ls` to show
-    // its output, it excludes the thing we just installed because that flag.
-    // The summary output we get should be unfiltered, showing everything
-    // installed, so we clear these options before calling `ls`.
-    npm.config.set('production', false)
-    npm.config.set('dev', false)
-    npm.config.set('only', '')
-    npm.config.set('also', '')
-    ls.fromTree(self.where, tree, addedOrMoved, false, andIgnoreErrors(cb))
-  }))
 }
 
 Installer.prototype.debugActions = function (name, actionListName, cb) {
@@ -720,21 +966,43 @@ Installer.prototype.debugActions = function (name, actionListName, cb) {
 // to define the arguments for use by chain before the property exists yet.
 Installer.prototype.debugTree = function (name, treeName, cb) {
   validate('SSF', arguments)
-  log.silly(name, this.prettify(this[treeName]).trim())
+  log.silly(name, this.archyDebugTree(this[treeName]).trim())
   cb()
 }
 
-Installer.prototype.prettify = function (tree) {
+Installer.prototype.archyDebugTree = function (tree) {
   validate('O', arguments)
-  var seen = {}
+  var seen = new Set()
   function byName (aa, bb) {
     return packageId(aa).localeCompare(packageId(bb))
   }
   function expandTree (tree) {
-    seen[tree.path] = true
+    seen.add(tree)
     return {
       label: packageId(tree),
-      nodes: tree.children.filter(function (tree) { return !seen[tree.path] }).sort(byName).map(expandTree)
+      nodes: tree.children.filter((tree) => { return !seen.has(tree) && !tree.removed }).sort(byName).map(expandTree)
+    }
+  }
+  return archy(expandTree(tree), '', { unicode: npm.config.get('unicode') })
+}
+
+Installer.prototype.debugLogicalTree = function (name, treeName, cb) {
+  validate('SSF', arguments)
+  this[treeName] && log.silly(name, this.archyDebugLogicalTree(this[treeName]).trim())
+  cb()
+}
+
+Installer.prototype.archyDebugLogicalTree = function (tree) {
+  validate('O', arguments)
+  var seen = new Set()
+  function byName (aa, bb) {
+    return packageId(aa).localeCompare(packageId(bb))
+  }
+  function expandTree (tree) {
+    seen.add(tree)
+    return {
+      label: packageId(tree),
+      nodes: tree.requires.filter((tree) => { return !seen.has(tree) && !tree.removed }).sort(byName).map(expandTree)
     }
   }
   return archy(expandTree(tree), '', { unicode: npm.config.get('unicode') })
